#!/usr/bin/env python

"""Tests for `yenta` package."""

import pytest
import json

from pathlib import Path
from click.testing import CliRunner

from yenta import cli
from yenta.pipeline import PipelineResult


@pytest.fixture
def response():
    """Sample pytest fixture.

    See more at: http://doc.pytest.org/en/latest/fixture.html
    """
    # import requests
    # return requests.get('https://github.com/audreyr/cookiecutter-pypackage')


def test_content(response):
    """Sample pytest test function with the pytest fixture as an argument."""
    # from bs4 import BeautifulSoup
    # assert 'GitHub' in BeautifulSoup(response.content).title.string


def test_command_line_interface():
    """Test the CLI."""
    runner = CliRunner()
    result = runner.invoke(cli.yenta)
    assert result.exit_code == 0
    for cmd in ['dump-task-graph', 'list-tasks', 'rm', 'run', 'show-config', 'task-info']:
        assert cmd in result.output
#
#     # Path('pipeline.json').unlink()
#     # help_result = runner.invoke(cli.yenta, ['--help'])
#     # assert help_result.exit_code == 0
#     # assert '--help  Show this message and exit.' in help_result.output


def test_list_tasks():

    runner = CliRunner()
    entry_point = 'tests/sample_pipelines/sample_pipeline_1.py'
    pipeline_store = 'tests/sample_pipelines/sample_pipeline_1.json'
    if Path(pipeline_store).exists():
        Path(pipeline_store).unlink()

    result = runner.invoke(cli.yenta, ['--entry-point', entry_point, '--pipeline', pipeline_store, 'list-tasks'])

    output_lines = result.output.split('\n')
    assert result.exit_code == 0
    assert output_lines[0] == 'The following tasks are available:'
    assert output_lines[1] == '[ ] foo'
    assert output_lines[2] == '[ ] bar'
    assert output_lines[3] == '[ ] baz'

    if Path(pipeline_store).exists():
        Path(pipeline_store).unlink()


def test_run_tasks():

    runner = CliRunner()
    entry_point = 'tests/sample_pipelines/sample_pipeline_1.py'
    pipeline_store = 'tests/sample_pipelines/sample_pipeline_1.json'
    if Path(pipeline_store).exists():
        Path(pipeline_store).unlink()

    result = runner.invoke(cli.yenta, ['--entry-point', entry_point, '--pipeline', pipeline_store, 'run'])

    assert result.exit_code == 0

    # bar errors, foo runs, baz never gets called
    output_lines = result.output.split('\n')
    assert output_lines[0] == '[\u2718] bar'
    assert output_lines[1] == 'hello from foo task'
    assert output_lines[2] == '[\u2714] foo'

    result = runner.invoke(cli.yenta, ['--entry-point', entry_point, '--pipeline', pipeline_store, 'run'])

    # bar still errors, foo recycles the old value, baz still not called
    output_lines = result.output.split('\n')
    assert output_lines[0] == '[\u2718] bar'
    assert output_lines[1] == '[\u2014] foo'

    result = runner.invoke(cli.yenta, ['--entry-point', entry_point, '--pipeline', pipeline_store, 'list-tasks'])

    output_lines = result.output.split('\n')
    assert output_lines[0] == 'The following tasks are available:'
    assert output_lines[1] == '[\u2714] foo'
    assert output_lines[2] == '[\u2718] bar'
    assert output_lines[3] == '[ ] baz'

    if Path(pipeline_store).exists():
        Path(pipeline_store).unlink()


def test_show_config():

    runner = CliRunner()
    entry_point = 'tests/sample_pipelines/sample_pipeline_1.py'
    pipeline_store = 'tests/sample_pipelines/sample_pipeline_1.json'
    if Path(pipeline_store).exists():
        Path(pipeline_store).unlink()

    result = runner.invoke(cli.yenta, ['--entry-point', entry_point, '--pipeline', pipeline_store, 'show-config'])

    output_lines = result.output.split('\n')
    assert output_lines[0] == 'Yenta is using the following configuration:'
    assert output_lines[1] == f'The entrypoint for Yenta is {entry_point}'
    assert output_lines[2] == f'Pipeline will be cached in {pipeline_store}'
    assert output_lines[3] == 'No log output configured'
    assert output_lines[4] == 'Tasks will be executed in the following order: bar, foo, baz'
    print(result.output)


def test_task_info():

    runner = CliRunner()
    entry_point = 'tests/sample_pipelines/sample_pipeline_1.py'
    pipeline_store = 'tests/sample_pipelines/sample_pipeline_1.json'
    if Path(pipeline_store).exists():
        Path(pipeline_store).unlink()

    result = runner.invoke(cli.yenta, ['--entry-point', entry_point,
                                       '--pipeline', pipeline_store,
                                       'task-info', 'nonexistent-task'])

    assert result.exit_code == 0
    assert result.output == 'Unknown task nonexistent-task specified.\n'

    result = runner.invoke(cli.yenta, ['--entry-point', entry_point,
                                       '--pipeline', pipeline_store,
                                       'run'])

    assert result.exit_code == 0
    assert Path(pipeline_store).exists()

    result = runner.invoke(cli.yenta, ['--entry-point', entry_point,
                                       '--pipeline', pipeline_store,
                                       'task-info', 'foo'])

    output_lines = result.output.split('\n')
    assert output_lines[0] == 'Information for task foo:'
    assert output_lines[1] == 'Dependencies: None'
    assert output_lines[2] == 'Previous status: \u2714'

    result = runner.invoke(cli.yenta, ['--entry-point', entry_point,
                                       '--pipeline', pipeline_store,
                                       'task-info', 'bar'])

    output_lines = result.output.split('\n')
    assert output_lines[0] == 'Information for task bar:'
    assert output_lines[1] == 'Dependencies: None'
    assert output_lines[2] == 'Previous status: \u2718 oh noes'

    result = runner.invoke(cli.yenta, ['--entry-point', entry_point,
                                       '--pipeline', pipeline_store,
                                       'task-info', 'baz'])

    output_lines = result.output.split('\n')
    assert output_lines[0] == 'Information for task baz:'
    assert output_lines[1] == 'Dependencies: foo, bar'
    assert output_lines[2] == 'Previous status: Did not run'

    if Path(pipeline_store).exists():
        Path(pipeline_store).unlink()


def test_rm_task():

    runner = CliRunner()
    entry_point = 'tests/sample_pipelines/sample_pipeline_1.py'
    pipeline_store = 'tests/sample_pipelines/sample_pipeline_1.json'
    if Path(pipeline_store).exists():
        Path(pipeline_store).unlink()

    result = runner.invoke(cli.yenta, ['--entry-point', entry_point,
                                       '--pipeline', pipeline_store,
                                       'run'])

    assert result.exit_code == 0
    assert Path(pipeline_store).exists()

    result = runner.invoke(cli.yenta, ['--entry-point', entry_point,
                                       '--pipeline', pipeline_store,
                                       'rm', 'foo'])

    assert result.exit_code == 0
    assert Path(pipeline_store).exists()

    with open(pipeline_store, 'r') as f:
        pipeline = PipelineResult(**json.load(f))

    with pytest.raises(KeyError) as ex:
        _ = pipeline.values('foo', 'whatever')

    assert 'foo' in str(ex.value)

    result = runner.invoke(cli.yenta, ['--entry-point', entry_point,
                                       '--pipeline', pipeline_store,
                                       'rm', 'nonexistent-task'])

    assert result.exit_code == 0
    assert result.output == 'Unknown task nonexistent-task specified.\n'

    if Path(pipeline_store).exists():
        Path(pipeline_store).unlink()


def test_dump_task_graph():

    runner = CliRunner()
    entry_point = 'tests/sample_pipelines/sample_pipeline_1.py'
    pipeline_store = 'tests/sample_pipelines/sample_pipeline_1.json'
    if Path(pipeline_store).exists():
        Path(pipeline_store).unlink()

    task_graph = 'tests/sample_pipelines/sample_task_graph.png'

    result = runner.invoke(cli.yenta, ['--entry-point', entry_point,
                                       '--pipeline', pipeline_store,
                                       'dump-task-graph', task_graph])

    assert result.exit_code == 0
    assert Path(task_graph).exists()

    Path(task_graph).unlink()



import json
import pytest
import networkx as nx

from datetime import datetime
from pathlib import Path

from yenta.config import settings
from yenta.tasks.Task import task
from yenta.pipeline import Pipeline, TaskResult, PipelineResult, InvalidTaskResultError
from yenta.values import Value
from yenta.artifacts import FileArtifact

settings.YENTA_JSON_STORE_PATH = Path('tests/tmp/pipeline.json')


def test_pipeline_creation():

    @task
    def foo(previous_results=None):
        pass

    @task
    def bar():
        pass

    @task(depends_on=['foo', 'bar'])
    def baz(previous_results):
        pass

    pipeline = Pipeline(foo, bar, baz)
    assert(pipeline.task_graph.has_edge('foo', 'baz'))
    assert(pipeline.task_graph.has_edge('bar', 'baz'))
    assert(not pipeline.task_graph.has_edge('foo', 'bar'))


def test_pipeline_with_cycles():

    @task(depends_on=['baz'])
    def foo(previous_results=None):
        pass

    @task(depends_on=['foo'])
    def bar():
        pass

    @task(depends_on=['bar'])
    def baz(previous_results):
        pass

    with pytest.raises(nx.NetworkXUnfeasible):
        pipeline = Pipeline(foo, bar, baz)


def test_run_pipeline_with_past_results():

    if settings.YENTA_JSON_STORE_PATH.exists():
        settings.YENTA_JSON_STORE_PATH.unlink()

    @task
    def foo(previous_results=None) -> TaskResult:
        return TaskResult({'x': Value(1)}, {})

    @task
    def bar():
        return TaskResult({'y': Value(2)}, {})

    @task(depends_on=['foo', 'bar'])
    def baz(previous_results: PipelineResult):

        x = previous_results.values('foo', 'x')
        y = previous_results.values('bar', 'y')

        result = x + y

        return TaskResult({'sum': Value(result)}, {})

    pipeline = Pipeline(foo, bar, baz)
    result = pipeline.run_pipeline()

    sum = result.values('baz', 'sum')
    assert(sum == 3)

    if settings.YENTA_JSON_STORE_PATH.exists():
        settings.YENTA_JSON_STORE_PATH.unlink()


def test_pipeline_run_with_explicit_params():

    if settings.YENTA_JSON_STORE_PATH.exists():
        settings.YENTA_JSON_STORE_PATH.unlink()

    @task
    def foo():
        return {'values': {'x': Value(1)}}

    @task
    def bar():
        return TaskResult({'y': Value(2)}, {})

    @task(depends_on=['foo', 'bar'])
    def baz(x: 'foo__values__x', y: 'bar__values__y'):

        result = x + y

        return TaskResult({'sum': Value(result)}, {})

    pipeline = Pipeline(foo, bar, baz)
    result = pipeline.run_pipeline()
    sum = result.values('baz', 'sum')

    assert (sum == 3)
    assert (pipeline._tasks_executed == {'foo', 'bar', 'baz'})
    assert (pipeline._tasks_reused == set())

    raw_1 = json.load(open(settings.YENTA_JSON_STORE_PATH, 'r'))
    t1 = settings.YENTA_JSON_STORE_PATH.stat().st_mtime

    result = pipeline.run_pipeline()
    sum = result.values('baz', 'sum')
    assert (sum == 3)
    assert (pipeline._tasks_reused == {'foo', 'bar', 'baz'})

    raw_2 = json.load(open(settings.YENTA_JSON_STORE_PATH, 'r'))
    t2 = settings.YENTA_JSON_STORE_PATH.stat().st_mtime

    assert(raw_1 == raw_2)
    assert(t1 != t2)

    if settings.YENTA_JSON_STORE_PATH.exists():
        settings.YENTA_JSON_STORE_PATH.unlink()


def test_pipeline_run_with_artifacts():

    foo_file = './foo.dat'
    bar_file = './bar.dat'
    baz_file = './baz.dat'

    if settings.YENTA_JSON_STORE_PATH.exists():
        settings.YENTA_JSON_STORE_PATH.unlink()

    @task
    def foo() -> TaskResult:
        with open(foo_file, 'w') as f:
            f.write('foo')
        return TaskResult({'x': 1},
                          {'foo_file': FileArtifact(foo_file, str(datetime.now()))})

    @task
    def bar():
        with open(bar_file, 'w') as f:
            f.write('bar')
        return TaskResult({'y': Value(2)},
                          {'bar_file': FileArtifact(bar_file, str(datetime.now()))})

    @task(depends_on=['foo', 'bar'])
    def baz(x: 'foo__values__x', y: 'bar__values__y',
            foo_artifact: 'foo__artifacts__foo_file', bar_artifact: 'bar__artifacts__bar_file'):
        sum_x_y = x + y
        with open(foo_artifact.location, 'r') as f:
            foo_data = f.read()
        with open(bar_artifact.location, 'r') as f:
            bar_data = f.read()
        with open(baz_file, 'w') as f:
            f.write(foo_data + bar_data)
        return TaskResult({'sum': Value(sum_x_y)},
                          {'baz_file': FileArtifact(baz_file, str(datetime.now()))})

    pipeline = Pipeline(foo, bar, baz)

    result = pipeline.run_pipeline()
    sum = result.values('baz', 'sum')
    assert (sum == 3)

    baz_artifact: FileArtifact = result.artifacts('baz', 'baz_file')

    with open(baz_artifact.location, 'r') as f:
        baz_data = f.read()

    assert(baz_data == 'foobar')

    Path(foo_file).unlink()
    Path(bar_file).unlink()
    Path(baz_file).unlink()

    if settings.YENTA_JSON_STORE_PATH.exists():
        settings.YENTA_JSON_STORE_PATH.unlink()


def test_pipeline_with_non_scalar_values():

    if settings.YENTA_JSON_STORE_PATH.exists():
        settings.YENTA_JSON_STORE_PATH.unlink()

    @task
    def foo() -> TaskResult:
        return TaskResult({'x': Value([1, 2, 3])})

    @task
    def bar():
        return TaskResult({'y': [4, 5, 6]})

    @task(depends_on=['foo', 'bar'])
    def baz(x: 'foo__values__x', y: 'bar__values__y'):
        sum_x_y = x + y
        return TaskResult({'result': Value(sum_x_y)})

    pipeline = Pipeline(foo, bar, baz)

    result = pipeline.run_pipeline()
    answer = result.values('baz', 'result')
    assert (answer == [1, 2, 3, 4, 5, 6])

    if settings.YENTA_JSON_STORE_PATH.exists():
        settings.YENTA_JSON_STORE_PATH.unlink()


def test_pipeline_run_with_selectors():

    if settings.YENTA_JSON_STORE_PATH.exists():
        settings.YENTA_JSON_STORE_PATH.unlink()

    @task
    def foo() -> TaskResult:
        return TaskResult({'x': Value([1, 2, 3])})

    @task
    def bar():
        return TaskResult({'y': [4, 5, 6]})

    def foo_x_selector(result: PipelineResult):
        return sum(result.values('foo', 'x'))

    def bar_y_selector(result: PipelineResult):
        return sum(result.values('bar', 'y'))

    @task(depends_on=['foo', 'bar'], selectors={'x': foo_x_selector, 'y': bar_y_selector})
    def baz(x, y):
        sum_x_y = x + y
        return TaskResult({'result': Value(sum_x_y)})

    pipeline = Pipeline(foo, bar, baz)

    result = pipeline.run_pipeline()
    answer = result.values('baz', 'result')
    assert (answer == 21)

    if settings.YENTA_JSON_STORE_PATH.exists():
        settings.YENTA_JSON_STORE_PATH.unlink()


def test_wrap_value():

    v = Value(1)
    assert v == TaskResult._wrap_as_value(v)
    assert v == TaskResult._wrap_as_value({'value': 1})
    assert v == TaskResult._wrap_as_value(1)

    with pytest.raises(ValueError) as ex:
        TaskResult._wrap_as_value({1})

    assert f'Can not wrap {set([1])} in a Value'


def test_wrap_task_output():

    t = TaskResult(values={'v': 1}, artifacts={})
    assert t == Pipeline._wrap_task_output({'values': {'v': 1}}, 'task_name')
    assert t == Pipeline._wrap_task_output(t, 'task_name')

    with pytest.raises(InvalidTaskResultError) as ex:
        Pipeline._wrap_task_output([1], 'task_name')

    assert "Task task_name returned invalid result of type <class 'list'>, " \
           "expected either a dict or a TaskResult" in str(ex.value)


def test_non_serializable_result():

    if settings.YENTA_JSON_STORE_PATH.exists():
        settings.YENTA_JSON_STORE_PATH.unlink()

    @task
    def foo() -> TaskResult:
        return TaskResult({'x': 1})

    # a task with a non-serializable result
    @task(depends_on=['foo'])
    def bar():
        return TaskResult({'y': Path('abc')})

    @task(depends_on=['bar'])
    def baz():
        return TaskResult({'z': 2})

    pipeline = Pipeline(foo, bar)

    # we'll have an exception that will prevent baz from running
    result = pipeline.run_pipeline()

    with open(settings.YENTA_JSON_STORE_PATH, 'r') as f:
        cached_result = json.load(f)
        assert result == PipelineResult(**cached_result)
        assert 'baz' not in cached_result



#!/usr/bin/env python3
"""Console script for yenta."""
import sys
import click
import configparser
import importlib.util
import more_itertools
import json
import networkx as nx

from dataclasses import asdict
from colorama import init, Fore, Style
from pathlib import Path
from yenta.config import settings
from yenta.config import logging
from yenta.pipeline.Pipeline import Pipeline, TaskStatus

import logging

logger = logging.getLogger(__name__)


def load_tasks(entry_file):
    spec = importlib.util.spec_from_file_location('main', entry_file)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    tasks = [func for _, func in module.__dict__.items()
             if callable(func) and hasattr(func, '_yenta_task')]

    return tasks


@click.group()
@click.option('--config-file', default=settings.YENTA_CONFIG_FILE, type=Path,
              help='The config file from which to read settings.')
@click.option('--pipeline', type=Path, help='The file to which the pipeline will be cached.')
@click.option('--entry-point', type=Path, help='The file containing the task definitions.')
@click.option('--log-file', type=Path, help='The file to which the logs should be written.')
def yenta(config_file, pipeline, entry_point, log_file):

    init()

    cf = configparser.ConfigParser()
    cf.read(config_file or settings.YENTA_CONFIG_FILE)
    if 'yenta' not in cf:
        cf['yenta'] = {}

    settings.YENTA_ENTRY_POINT = entry_point or \
                                 cf['yenta'].get('entry_point', None) or \
                                 settings.YENTA_ENTRY_POINT

    pipeline_file = cf['yenta'].get('pipeline_store', None)
    pipeline_path = Path(pipeline_file).resolve() if pipeline_file else None
    settings.YENTA_JSON_STORE_PATH = pipeline or \
                                     pipeline_path or \
                                     settings.YENTA_JSON_STORE_PATH
    conf_log_file = cf['yenta'].get('log_file', None)
    conf_log_path = Path(conf_log_file).resolve() if log_file else None
    settings.YENTA_LOG_FILE = log_file or \
                              conf_log_path or \
                              settings.YENTA_LOG_FILE


@yenta.command(help='List all available tasks.')
def list_tasks():

    tasks = load_tasks(settings.YENTA_ENTRY_POINT)
    pipeline_data = Pipeline.load_pipeline()

    print(Fore.WHITE + Style.BRIGHT + 'The following tasks are available:')
    for task in tasks:
        task_name = task.task_def.name
        task_result = pipeline_data.task_results.get(task_name, None)
        marker = ' '
        if task_result and task_result.status == TaskStatus.SUCCESS:
            marker = Fore.GREEN + u'\u2714' + Fore.WHITE
        elif task_result and task_result.status == TaskStatus.FAILURE:
            marker = Fore.RED + u'\u2718' + Fore.WHITE

        print(Fore.WHITE + Style.BRIGHT + f'[{marker}] {task_name}')


@yenta.command(help='Show the current configuration.')
def show_config():

    tasks = load_tasks(settings.YENTA_ENTRY_POINT)
    pipeline = Pipeline(*tasks)

    print(Fore.WHITE + Style.BRIGHT + 'Yenta is using the following configuration:')
    print('The entrypoint for Yenta is ' + Fore.GREEN + str(settings.YENTA_ENTRY_POINT) + Fore.WHITE)
    print('Pipeline will be cached in ' + Fore.GREEN + str(settings.YENTA_JSON_STORE_PATH) + Fore.WHITE)
    if settings.YENTA_LOG_FILE:
        print('Log output will be written to ' + Fore.GREEN + str(settings.YENTA_LOG_FILE) + Fore.WHITE)
    else:
        print('No log output configured')
    print('Tasks will be executed in the following order: ' + Fore.GREEN + ', '.join(pipeline.execution_order))


@yenta.command(help='Show information about a specific task.')
@click.argument('task-name')
def task_info(task_name):

    tasks = load_tasks(settings.YENTA_ENTRY_POINT)
    pipeline_data = Pipeline.load_pipeline()
    try:
        task = more_itertools.one(filter(lambda t: t.task_def.name == task_name, tasks))
        print(Fore.WHITE + Style.BRIGHT + 'Information for task ' + Fore.GREEN + task_name + Fore.WHITE + ':')
        deps = ', '.join(task.task_def.depends_on) if task.task_def.depends_on else 'None'
        print('Dependencies:', Fore.GREEN + deps + Fore.WHITE)
        task_result = pipeline_data.task_results.get(task_name, None)
        marker = 'Did not run'
        if task_result and task_result.status == TaskStatus.SUCCESS:
            marker = Fore.GREEN + u'\u2714' + Fore.WHITE
        elif task_result and task_result.status == TaskStatus.FAILURE:
            marker = Fore.RED + u'\u2718' + ' ' + task_result.error + Fore.WHITE
        print(f'Previous status: {marker}')
        if task_result and task_result.status == TaskStatus.SUCCESS:
            print('Previous result: ', task_result)
        else:
            print('Previous result: ' + Fore.GREEN + 'None' + Fore.WHITE)

    except ValueError:
        print(Fore.WHITE + Style.BRIGHT + 'Unknown task ' + Fore.RED + task_name + Fore.WHITE + ' specified.')


@yenta.command(help='Remove a task from the pipeline cache.')
@click.argument('task-name')
def rm(task_name):

    pipeline_data = Pipeline.load_pipeline()

    if task_name in pipeline_data.task_results:
        del pipeline_data.task_results[task_name]
    else:
        print(Fore.WHITE + Style.BRIGHT + 'Unknown task ' + Fore.RED + task_name + Fore.WHITE + ' specified.')
    if task_name in pipeline_data.task_inputs:
        del pipeline_data.task_inputs[task_name]

    with open(settings.YENTA_JSON_STORE_PATH, 'w') as f:
        json.dump(asdict(pipeline_data), f, indent=4)


@yenta.command(help='Dump the task graph to a file; requires Matplotlib.')
@click.argument('filename', type=click.Path())
def dump_task_graph(filename):

    try:
        import matplotlib.pyplot as plt
        tasks = load_tasks(settings.YENTA_ENTRY_POINT)
        pipeline = Pipeline(*tasks)
        nx.draw_networkx(pipeline.task_graph)
        plt.savefig(filename)
    except ImportError as ex:
        print(Fore.WHITE + Style.BRIGHT + f'Matplotlib must be installed to dump the task graph: {ex}')


@yenta.command(help='Run the pipeline.')
@click.option('--up-to', help='Optionally run the pipeline up to and including a given task.')
@click.option('--force-rerun', '-f', multiple=True, default=[], help='Force specified tasks to rerun.')
def run(up_to=None, force_rerun=None):

    logger.info('Running the pipeline')
    tasks = load_tasks(settings.YENTA_ENTRY_POINT)
    pipeline = Pipeline(*tasks)
    result = pipeline.run_pipeline(up_to, force_rerun)


if __name__ == "__main__":
    sys.exit(yenta())  # pragma: no cover



import io
import json
import networkx as nx
import logging
import shutil
import tempfile

from dataclasses import dataclass, field, asdict
from pathlib import Path
from typing import Dict, List, Union, Any, Callable
from enum import Enum
from more_itertools import split_after
from colorama import Fore, Style

from yenta.config import settings
from yenta.tasks.Task import TaskDef, ParameterType, ResultSpec
from yenta.artifacts.Artifact import Artifact
from yenta.values.Value import Value


logger = logging.getLogger(__name__)


class InvalidTaskResultError(Exception):
    pass


class InvalidParameterError(Exception):
    pass


class TaskStatus(str, Enum):

    SUCCESS = 'success'
    FAILURE = 'failure'


@dataclass
class TaskResult:
    """ Holds the result of a specific task execution """

    values: Dict[str, Union[Value, List, str, int, float, bool]] = field(default_factory=dict)
    """ A dictionary whose keys are value names and whose values are... values."""

    artifacts: Dict[str, Artifact] = field(default_factory=dict)
    """ A dictionary whose keys are artifact names and whose values are Artifacts."""

    status: TaskStatus = None
    """ Whether the task succeeded or failed."""

    error: str = None
    """ Error message associated with task failure."""

    @staticmethod
    def _wrap_as_value(v: Union[Value, dict, str, int, float, bool, list]) -> Value:
        """
        Wraps a value resulting from a task execution in a Value.
        :param v: the value to be wrapped
        :return: Value(v)
        """
        if isinstance(v, Value):
            return v
        elif isinstance(v, dict):
            return Value(**v)
        elif isinstance(v, str) or isinstance(v, int) or isinstance(v, float) or isinstance(v, bool) or \
                isinstance(v, list) or v is None:
            return Value(v)
        else:
            raise ValueError(f'Can not wrap {v} in a Value')

    def __post_init__(self):
        """
        Deserialize nested values and artifacts into Value and Artifact objects
        :return:
        """

        values = {k: self._wrap_as_value(v) for k, v in self.values.items() if not isinstance(v, Value)}
        artifacts = {k: Artifact(**v) for k, v in self.artifacts.items() if not isinstance(v, Artifact)}

        self.values.update(values)
        self.artifacts.update(artifacts)


@dataclass
class PipelineResult:
    """ Holds the intermediate results of a step in the pipeline, where the keys of the dicts
        are the names of the tasks that have been executed and the values are TaskResults"""

    task_results: Dict[str, TaskResult] = field(default_factory=dict)
    """ A dictionary whose keys are task names and whose values are the results of that task execution."""

    task_inputs: Dict[str, 'PipelineResult'] = field(default_factory=dict)
    """ A dictionary whose keys are task names and whose values are the inputs used in executing that task."""

    def __post_init__(self):
        """
        Deserialize subcomponents of the PipelineResult into task results and task inputs
        :return: None
        """

        task_results = {k: TaskResult(**v) for k, v in self.task_results.items() if not isinstance(v, TaskResult)}
        task_inputs = {k: PipelineResult(**v) for k, v in self.task_inputs.items() if not isinstance(v, PipelineResult)}

        self.task_results.update(task_results)
        self.task_inputs.update(task_inputs)

    def values(self, task_name: str, value_name: str):
        """ Return the value named `value_name` that was produced by task `task_name`.

        :param str task_name: The name of the task
        :param str value_name: The name of the value
        :return: the unwrapped value produced by the task
        :rtype: Union[list, int, bool, float, str]
        """
        return self.task_results[task_name].values[value_name].value

    def artifacts(self, task_name: str, artifact_name: str):
        """ Return the artifact names `artifact_name` that was produced by the task `task_name`.

        :param str task_name: The name of the task
        :param str artifact_name: The name of the artifact
        :return: The artifact produced by the task
        :rtype: Artifact
        """
        return self.task_results[task_name].artifacts[artifact_name]

    def from_spec(self, spec: ResultSpec):
        """ Return either the value or the artifact of a given task, as computed by
            a ResultSpec. Delegates the actual work to the `value` and `artifacts` functions.

        :param ResultSpec spec: The result spec
        :return: either the value or the artifact computed from the spec
        """
        func = getattr(self, spec.result_type)
        return func(spec.result_task_name, spec.result_var_name)


class Pipeline:

    def __init__(self, *tasks):

        self._tasks = tasks
        self.task_graph = nx.DiGraph()
        self.execution_order = []

        self.build_task_graph()

        self._tasks_executed = set()
        self._tasks_reused = set()

    def build_task_graph(self) -> None:
        """ Construct the task graph for the pipeline

        :return: None
        """
        logger.debug('Building task graph')
        for task in self._tasks:
            self.task_graph.add_node(task.task_def.name, task=task)
            for dependency in (task.task_def.depends_on or []):
                self.task_graph.add_edge(dependency, task.task_def.name)

        logger.debug('Computing execution order')
        try:
            self.execution_order = list(nx.algorithms.dag.lexicographical_topological_sort(self.task_graph))
        except nx.NetworkXUnfeasible as ex:
            print(Fore.RED + 'Unable to build execution graph because pipeline contains cyclic dependencies.')
            raise ex

    @staticmethod
    def _wrap_task_output(raw_output: Union[dict, TaskResult], task_name: str) -> TaskResult:
        """ Wrap the raw output of a task in a TaskResult.

        :param Union[dict, TaskResult] raw_output: The raw output of a task.
        :param task_name: The name of the task.
        :return: A TaskResult containing the output
        :rtype: TaskResult
        """

        if isinstance(raw_output, dict):
            output: TaskResult = TaskResult(**raw_output)
        elif isinstance(raw_output, TaskResult):
            output = raw_output
        else:
            raise InvalidTaskResultError(f'Task {task_name} returned invalid result of type {type(raw_output)}, '
                                         f'expected either a dict or a TaskResult')

        return output

    @staticmethod
    def build_args_dict(task, args: PipelineResult) -> Dict[str, Any]:
        """ Build the args dictionary for executing a task.

        :param task: The task itself, which has a `task_def` attached to it.
        :param PipelineResult args: The results of the pipeline up to this point
        :return: A dictionary whose keys correspond to the arguments expected by
                the task to be executed, and whose values are the values to be
                passed in.
        :rtype: Dict[str, Any]
        """

        logger.debug('Building args dictionary')
        task_def: TaskDef = task.task_def
        args_dict = {}

        for spec in task_def.param_specs:
            if spec.param_type == ParameterType.PIPELINE_RESULTS:
                args_dict[spec.param_name] = args
            elif spec.param_type == ParameterType.EXPLICIT:
                if spec.selector:
                    args_dict[spec.param_name] = spec.selector(args)
                elif spec.result_spec:
                    args_dict[spec.param_name] = args.from_spec(spec.result_spec)

        return args_dict

    def invoke_task(self, task, **kwargs) -> TaskResult:
        """ Call the function that represents the task with the supplied kwargs.

        :param Callable task: The task function.
        :param dict kwargs: The arguments obtained from `build_args`.
        :return: The task result
        :rtype: TaskResult
        """

        output = task(**kwargs)
        return self._wrap_task_output(output, task.task_def.name)

    @staticmethod
    def merge_pipeline_results(res1: PipelineResult, res2: PipelineResult) -> PipelineResult:
        """ Combine two different pipeline results. If they share keys,
            the results of the second pipeline will overwrite those of
            the first.

        :param PipelineResult res1: The first result.
        :param PipelineResult res2: The second result.
        :return: The merged result.
        :rtype: PipelineResult
        """

        return PipelineResult(task_results={**res1.task_results, **res2.task_results},
                              task_inputs={**res1.task_inputs, **res2.task_inputs})

    @staticmethod
    def cache_pipeline_result(cache_file: io.BufferedRandom, result: PipelineResult):
        """ Write the pipeline results to a file.

        :param Path cache_file: The file to which the results should be written.
        :param PipelineResult result: The results.
        :return: None
        """

        try:
            result = json.dumps(asdict(result), indent=4)
            cache_file.truncate(0)
            cache_file.flush()
            cache_file.write(result.encode('utf-8'))
            cache_file.flush()
        except json.JSONDecodeError as ex:
            logger.error(f'Failed to encode pipeline result: {ex}')
            raise InvalidTaskResultError(str(ex))

    @staticmethod
    def load_pipeline() -> PipelineResult:
        """ Load a pipeline from file.

        :return: The pipeline.
        :rtype: PipelineResult
        """
        logger.debug(f'Loading pipeline from {settings.YENTA_JSON_STORE_PATH}')
        if settings.YENTA_JSON_STORE_PATH.exists():
            with open(settings.YENTA_JSON_STORE_PATH, 'r') as f:
                pipeline = PipelineResult(**json.load(f))
        else:
            pipeline = PipelineResult()
        return pipeline

    @staticmethod
    def reuse_inputs(task_name: str, previous_result: PipelineResult, args: PipelineResult) -> bool:
        """ Determine whether inputs from the previous instance of this task should be reused
            or whether the task should be executed again.

        :param str task_name: The name of the task.
        :param PipelineResult previous_result: The previous pipeline result.
        :param PipelineResult args: The arguments with which this task is being called.
        :return: True or False
        :rtype: bool
        """
        previous_inputs = previous_result.task_inputs.get(task_name, None)
        if previous_inputs and previous_result.task_results.get(task_name).status == TaskStatus.SUCCESS:
            return previous_inputs == args

        return False

    def run_pipeline(self, up_to: str = None, force_rerun: List[str] = None) -> PipelineResult:
        """ Execute the tasks in the pipeline.

        :param str up_to: If supplied, execute the pipeline only up to this task.
        :param List[str] force_rerun: Optionally force the listed tasks to be executed.
        :return: The final pipeline state.
        :rtype: PipelineResult
        """

        previous_result: PipelineResult = self.load_pipeline()
        result = PipelineResult()
        self._tasks_reused.clear()
        self._tasks_executed.clear()

        with tempfile.NamedTemporaryFile(delete=False) as cache:

            for task_name in list(split_after(self.execution_order, lambda x: x == up_to))[0]:
                logger.debug(f'Starting executions of {task_name}')
                task = self.task_graph.nodes[task_name]['task']
                args = PipelineResult()
                dependencies_succeeded = True
                for dependency in (task.task_def.depends_on or []):
                    args.task_results[dependency] = result.task_results[dependency]
                    if result.task_results[dependency].status == TaskStatus.FAILURE:
                        dependencies_succeeded = False

                if dependencies_succeeded:
                    if task.task_def.pure and task_name not in (force_rerun or []) and \
                            self.reuse_inputs(task_name, previous_result, args):
                        logger.debug(f'Reusing previous results of {task_name}')
                        self._tasks_reused.add(task_name)
                        output = previous_result.task_results[task_name]
                        marker = Fore.YELLOW + u'\u2014' + Fore.WHITE
                    else:
                        args_dict = self.build_args_dict(task, args)
                        try:
                            logger.debug(f'Calling function to execute {task_name}')
                            output = self.invoke_task(task, **args_dict)
                            output.status = TaskStatus.SUCCESS
                            marker = Fore.GREEN + u'\u2714' + Fore.WHITE
                            self._tasks_executed.add(task_name)
                        except Exception as ex:
                            logger.error(f'Caught exception executing {task_name}: {ex}')
                            output = TaskResult(status=TaskStatus.FAILURE, error=str(ex))
                            marker = Fore.RED + u'\u2718' + Fore.WHITE

                    print(Fore.WHITE + Style.BRIGHT + f'[{marker}] {task_name}')

                    result.task_results[task_name] = output
                    result.task_inputs[task_name] = args

                    self.cache_pipeline_result(cache, self.merge_pipeline_results(previous_result, result))

            cache.seek(0)
            full_results = cache.read().decode('utf-8').strip('\x00')
            with open(settings.YENTA_JSON_STORE_PATH, 'w') as f:
                f.write(full_results)

        return result
